# -*- coding: utf-8 -*-
"""
Created on Fri Nov 18 17:31:34 2016

@author: steven
"""

import genmechanics
import genmechanics.linkages as linkages
import genmechanics.loads as loads
import numpy as npy
import scipy.linalg as linalg

F=100
L=0.1

ground=genmechanics.Part('ground')
shaft1=genmechanics.Part('shaft1')

bearing=genmechanics.linkages.RevoluteLinkage(ground,shaft1,[0,0,0],[0,0,0],'bearing')

mech=genmechanics.Mechanism([bearing],ground)

load1=loads.KnownMechanicalLoad(shaft1,[0,L,0],[0,0,0],[0,0,F],[0,0,0],'load')
load2=loads.UnknownMechanicalLoad(shaft1,[0,0,0],[0,0,0],[],[0],'output_torque')

r=mech.StaticAnalysis([load1],[load2])

for (l,d),v in r.items():
    print(l.name,d,v)


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 16 11:06:30 2016

@author: steven
"""

import numpy as npy
import networkx as nx
from genmechanics import geometry,tools
from scipy import linalg
from scipy.optimize import fsolve
import webbrowser
import os
from jinja2 import Environment, PackageLoader, select_autoescape

class ModelError(Exception):
    def __init__(self,message):
        self.message=message

    def __str__(self):
        return 'Model Error: '+self.message

class Part:
    def __init__(self,name=''):
        self.name=name
        
        
class Mechanism:
    def __init__(self,linkages,ground,imposed_speeds,known_static_loads,unknown_static_loads,name=''):
        self.linkages=linkages
        self.ground=ground
        self.name=name
        self.imposed_speeds=imposed_speeds
        self.known_static_loads=known_static_loads
        self.unknown_static_loads=unknown_static_loads
        
        self._utd_kinematic_results=False
        self._utd_static_results=False
        
        
    def _get_parts(self):
        parts=[]
        for linkage in self.linkages:
            for part in [linkage.part1,linkage.part2]:
                if not part in parts:
                    if part!=self.ground:
                        parts.append(part)
        return parts

    parts=property(_get_parts)

    def _get_graph(self):
        G=nx.Graph()
        G.add_nodes_from(self.parts)
        for linkage in self.linkages:
            G.add_node(linkage)
            G.add_edge(linkage,linkage.part1)
            G.add_edge(linkage,linkage.part2)
        return G

    graph=property(_get_graph)

    def _get_holonomic_graph(self):
        G=nx.Graph()
        G.add_nodes_from(self.parts)
        for linkage in self.linkages:
            if linkage.holonomic:
                G.add_node(linkage)
                G.add_edge(linkage,linkage.part1)
                G.add_edge(linkage,linkage.part2)
        return G

    holonomic_graph=property(_get_holonomic_graph)        

    def DrawGraph(self,n_state=None):
        """
        Draw graph with tulip
        """
        import tulip
        import tulipgui
        # Creating tulip graph
        G=tulip.tlp.newGraph()
        nodes={}
        for part in self.parts:
            nodes[part]=G.addNode()
        nodes[self.ground]=G.addNode()
        for linkage in self.linkages:
            nodes[linkage]=G.addNode()          
            for part in [linkage.part1,linkage.part2]:
                G.addEdge(nodes[part],nodes[linkage])

        
        viewLayout = G.getLayoutProperty("viewLayout")
        
        # Apply an FM^3 layout on it
        fm3pParams = tulip.tlp.getDefaultPluginParameters("FM^3 (OGDF)", G)
        fm3pParams["Unit edge length"] = 7
        G.applyLayoutAlgorithm("FM^3 (OGDF)", viewLayout, fm3pParams)

        viewLabel = G.getStringProperty("viewLabel")
        viewColor = G.getColorProperty("viewColor")
        viewShape = G.getIntegerProperty("viewShape")

        for comp,node in nodes.items():
            try:
                viewLabel[node] = comp.name

            except AttributeError:
                viewLabel[node] = comp.__class__.__name__
                viewColor[node]=tulip.tlp.Color.Gray
                viewShape[node]=tulip.tlp.NodeShape.Cube
                
        nodeLinkView = tulipgui.tlpgui.createNodeLinkDiagramView(G)

        raise TypeError
        
    def DrawVisJSGraph(self):
        s="""<html>
        <head>
        <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.20.0/vis.min.js"></script>
        <link href="https://cdnjs.cloudflare.com/ajax/libs/vis/4.20.0/vis.min.css" rel="stylesheet" type="text/css" />

        <style type="text/css">
            #mynetwork {
                border: 1px solid lightgray;
            }
        </style>
    </head>
    <body>
    <div id="mynetwork"></div>
    
    <script type="text/javascript">
    var nodes = new vis.DataSet([\n"""
        index={}
        for ipart,part in enumerate(self.parts+[self.ground]):
            index[part]=ipart
            s+="{{id: {}, label: '{}'}},\n".format(ipart,part.name)
#        s+=']);\n'
        n=len(self.parts)+1
#        index[self.ground]=n
#        n+=1
        for il,linkage in enumerate(self.linkages):
            index[linkage]=n+il
            s+="{{id: {}, label: '{}'}},\n".format(n+il,linkage.name)
        s+=']);\n'

        s+="var edges = new vis.DataSet(["
        for linkage in self.linkages:            
            s+='{{from: {}, to: {}}},\n'.format(index[linkage],index[linkage.part1])
            s+='{{from: {}, to: {}}},\n'.format(index[linkage],index[linkage.part2])
        s+=']);'

        s+="""
    // create a network
    var container = document.getElementById('mynetwork');

    // provide the data in the vis format
    var data = {
        nodes: nodes,
        edges: edges
    };
    var options = {};

    // initialize your network!
    var network = new vis.Network(container, data, options);
</script>
</body>
</html>"""

        with open('gm_graph_viz.html','w') as file:
            file.write(s)
        
        webbrowser.open('file://' + os.path.realpath('gm_graph_viz.html'))

    def DrawPowerGraph(self):
        """
        Draw graph with tulip
        """
        import matplotlib.pyplot as plt

        # Creating tulip graph
        G=nx.Graph()
#        nodes={}
#        edges={}
        widths=[]
        labels={}
        edges=[]
        for part in self.parts:
            G.add_node(part)
            
            labels[part]=part.name
#        nodes[self.ground]=G.addNode()
        for linkage in self.linkages:
            G.add_node(linkage)          
            labels[linkage]=linkage.name
#            for part in [linkage.part1,linkage.part2]:
            G.add_edge(linkage,linkage.part1)
            widths.append(abs(self.TransmittedLinkagePower(linkage,0)))
            edges.append((linkage,linkage.part1))
            G.add_edge(linkage,linkage.part2)
            widths.append(abs(self.TransmittedLinkagePower(linkage,1)))
            edges.append((linkage,linkage.part2))
            
        for load in self.unknown_static_loads+self.known_static_loads:
            G.add_node(load)
            G.add_edge(load,load.part)
            widths.append(abs(self.LoadPower(load)))
            labels[load]=load.name
            edges.append((load,load.part))
        max_widths=max(widths)
#        print(widths)
        widths=[6*w/max_widths for w in widths]                       
#        edges[linkage,part]=e
        plt.figure()
        pos=nx.spring_layout(G)
        nx.draw_networkx_nodes(G,pos,nodelist=self.linkages,node_color='grey')
        nx.draw_networkx_nodes(G,pos,nodelist=self.parts)
        nx.draw_networkx_nodes(G,pos,nodelist=self.unknown_static_loads,node_color='red')
        nx.draw_networkx_nodes(G,pos,nodelist=self.known_static_loads,node_color='green')
        nx.draw_networkx_nodes(G,pos,nodelist=self.parts,node_color='cyan')
        nx.draw_networkx_labels(G,pos,labels)
        nx.draw_networkx_edges(G,pos,edges,width=widths,edge_color='blue')
        nx.draw_networkx_edges(G,pos)
#        nx.draw_networkx_labels(G,pos,labels)

    def ChangeImposedSpeeds(self,imposed_speeds):
        self.imposed_speeds=imposed_speeds
        self._utd_kinematic_results=False
        self._utd_static_results=False# Change of speeds might change resistant forces: update needed

    def ChangeLoads(self,known_static_loads,unknown_static_loads):
        self.known_static_loads=known_static_loads
        self.unknown_static_loads=unknown_static_loads
        self._utd_static_results=False
        
        
    def Speeds(self,position,part_ref,part):
        """
        Speeds from point belonging to part with part_ref as reference
        """
        q=self.kinematic_vector#force kdof computation
        path=nx.shortest_path(self.holonomic_graph,part_ref,part)
        V=npy.zeros((3,self.n_kdof))
        W=npy.zeros((3,self.n_kdof))
        for il,linkage2 in enumerate(path):
            if not linkage2.__class__.__name__=='Part':
                try:
                    if path[il+1]==linkage2.part2:
                        side=1                            
                    else:
                        side=-1
                except IndexError:
                    # linkage is last element of list
                    if path[il-1]==linkage2.part1:
                        side=1     
                    else:
                        side=-1
                # It's really a linkage
                P=geometry.Euler2TransferMatrix(*linkage2.euler_angles)            
                u=linkage2.position
                uprime=u-position
                L=geometry.CrossProductMatrix(uprime)
                Ve=npy.dot(L,npy.dot(P,linkage2.kinematic_matrix[:3,:]))+npy.dot(P,linkage2.kinematic_matrix[3:,:])
                We=npy.dot(P,linkage2.kinematic_matrix[:3,:])
                for indof,ndof in enumerate(self.kdof[linkage2]):
                    V[:,ndof]+=side*Ve[:,indof] 
                    W[:,ndof]+=side*We[:,indof] 
                
        return npy.hstack((npy.dot(W,q).flatten(),npy.dot(V,q).flatten()))

    
    def LocalLinkageSpeeds(self,linkage):
        """
        :returns: relative speed of linkage in local linkage coordinate system
        """

        try:
            # holonomic linkage
            r=self.kinematic_results[linkage]# results
            vr=npy.array([r[i] for i in range(linkage.n_kinematic_unknowns)])#vector of results
            return npy.dot(linkage.kinematic_matrix,vr).flatten()
        except:
            s=self.Speed(linkage.position,self.ground,linkage.part1)
            v=npy.dot(linkage.P,s[:3])
            w=npy.dot(linkage.P,s[3:])
#            return npy.dot(linkage.P,s)
            return npy.hstack((w,v)).flatten()
    
    def LocalLinkageForces(self,linkage,num_part):
        """
        :returns :Local forces of linkage on part given by its number (0 for part1, 1 for part2)
        """
        r=self.static_results[linkage]# results
        vr=npy.array([r[i] for i in range(linkage.n_static_unknowns)])#vector of results
        if num_part==0:
            return npy.dot(linkage.static_matrix1,vr)
        else:
            return npy.dot(linkage.static_matrix2,vr)

    def GlobalLinkageForces(self,linkage,num_part):
        P=geometry.Euler2TransferMatrix(*linkage.euler_angles)
        lf=self.LocalLinkageForces(linkage,num_part)
        F=npy.zeros(6)
        F[:3]=npy.dot(P,lf[:3])
        F[3:]=npy.dot(P,lf[3:])
        return F
    
    def LocalLoadForces(self,load):
        try:
            F=npy.zeros(6)
            F[:3]=load.forces
            F[3:]=load.torques
            return F
        except AttributeError:       
            r=self.static_results[load]# results
            vr=npy.array([r[i] for i in range(load.n_static_unknowns)])#vector of results
            return npy.dot(load.static_matrix,vr)

    
    def GlobalLoadForces(self,load):
        f=self.LocalLoadForces(load)
#        print(f)
        F=npy.zeros(6)
        F[:3]=npy.dot(load.P,f[:3]).flatten()
#        print(npy.dot(load.P,f[3:]).flatten())
        F[3:]=npy.dot(load.P,f[3:]).flatten()
        return F

    
    def LinkagePowerLosses(self,linkage):
        if linkage.holonomic:
            tf=self.LocalLinkageForces(linkage,0)
            ft=npy.hstack((tf[3:],tf[:3]))
            return npy.dot(self.LocalLinkageSpeeds(linkage),ft)
        else:
            d=self.GlobalLinkageForces(linkage,0)+self.GlobalLinkageForces(linkage,1)
            df=d[:3]
            dt=d[3:]
            s=-self.Speeds(linkage.position,self.ground,linkage.part1)
            w=s[:3]
            v=s[3:]
#            print(df,dt,w,v)
            return npy.dot(df,v)+npy.dot(dt,w)

            
    def LoadPower(self,load):
        s=self.Speeds(load.position,self.ground,load.part)
        f=self.GlobalLoadForces(load)
        P=npy.dot(s[3:],f[:3])
        P+=npy.dot(f[3:],s[:3])
        return P
    
    def TransmittedLinkagePower(self,linkage,num_part):
        if num_part==1:
            s=self.Speeds(linkage.position,self.ground,linkage.part2)
        else:
            s=self.Speeds(linkage.position,self.ground,linkage.part1)            
        f=self.GlobalLinkageForces(linkage,num_part)
        P=npy.dot(s[3:],f[:3])
        P+=npy.dot(f[3:],s[:3])
        return P
        
    
    def _get_static_results(self):
        if not self._utd_static_results:
            self._static_results=self._StaticAnalysis()
            self._utd_static_results=True
        return self._static_results

    static_results=property(_get_static_results)
        
    def _get_kinematic_results(self):
        if not self._utd_kinematic_results:
            self._kinematic_results,self._kinematic_vector=self._KinematicAnalysis()
            self._utd_kinematic_results=True
        return self._kinematic_results

    kinematic_results=property(_get_kinematic_results)
    
    def _get_kinematic_vector(self):
        if not self._utd_kinematic_results:
            self._kinematic_results,self._kinematic_vector=self._KinematicAnalysis()
            self._utd_kinematic_results=True
        return self._kinematic_vector

    kinematic_vector=property(_get_kinematic_vector)

        
    def _StaticAnalysis(self):
        # Static parametrisation
        self.sdof={}
        self.n_sdof=0
        kinematic_analysis_required=False
        for linkage in self.linkages:
            self.sdof[linkage]=list(range(self.n_sdof,self.n_sdof+linkage.n_static_unknowns))
            self.n_sdof+=linkage.n_static_unknowns
            if linkage.static_require_kinematic:
                kinematic_analysis_required=True
              
                  
        uloads_parts={}
        for load in self.unknown_static_loads:
            load_unknowns=load.static_matrix.shape[1]
            self.sdof[load]=list(range(self.n_sdof,self.n_sdof+load_unknowns))
            self.n_sdof+=load_unknowns
            if load.static_require_kinematic:
                kinematic_analysis_required=True
            
            # Loading sorting by part
            try:
                uloads_parts[load.part].append(load)
            except:
                uloads_parts[load.part]=[load]

        # Force kinematic computation if required
        if kinematic_analysis_required:
            self.kinematic_results

        # Knowns loads sorting by parts
        loads_parts={}
        for load in self.known_static_loads:
            try:
                loads_parts[load.part].append(load)
            except:
                loads_parts[load.part]=[load]
                
                
        
        lparts=len(self.parts)
        M=npy.zeros((6*lparts,self.n_sdof))
        K=npy.zeros((6*lparts,self.n_sdof))
        F=npy.zeros(6*lparts)
        q=npy.zeros(self.n_sdof)
        nonlinear_eq={}

        # Occurence matrix assembly
        for ip,part in enumerate(self.parts):
            # linkage contribution to the LHS
            for linkage in self.graph[part].keys():
                P=geometry.Euler2TransferMatrix(*linkage.euler_angles)            
                u=linkage.position
                uprime=u
                L=geometry.CrossProductMatrix(uprime)
                if part==linkage.part1:
                    static_matrix=linkage.static_matrix1
                else:
                    static_matrix=linkage.static_matrix2
                    
                Me1=npy.abs(npy.dot(P,static_matrix[:3,:]))>1e-10
                Me2=npy.abs(npy.dot(L,npy.dot(P,static_matrix[:3,:]))+npy.dot(P,static_matrix[3:,:]))>1e-10
                Me=npy.vstack([Me1,Me2])
                for indof,ndof in enumerate(self.sdof[linkage]):
                    M[ip*6:(ip+1)*6,ndof]+=Me[:,indof]

                
                Ke1=npy.dot(P,static_matrix[:3,:])
                Ke2=npy.dot(L,npy.dot(P,static_matrix[:3,:]))+npy.dot(P,static_matrix[3:,:])
                Ke=npy.vstack([Ke1,Ke2])
                for indof,ndof in enumerate(self.sdof[linkage]):
                    K[ip*6:(ip+1)*6,ndof]+=Ke[:,indof]

                    
            # Unknowns loads contribution to the LHS
            try:
                uloads=uloads_parts[part]
            except:
                uloads=[]                
            for load in uloads:
                P=geometry.Euler2TransferMatrix(*load.euler_angles)            
                u=load.position
                uprime=u
                L=geometry.CrossProductMatrix(uprime)
                Me1=npy.abs(npy.dot(P,load.static_matrix[:3,:]))>1e-10
                Me2=npy.abs(npy.dot(L,npy.dot(P,load.static_matrix[:3,:]))+npy.dot(P,load.static_matrix[3:,:]))>1e-10
                Me=npy.vstack([Me1,Me2])
                for indof,ndof in enumerate(self.sdof[load]):
                    M[ip*6:(ip+1)*6,ndof]+=Me[:,indof]
                
                Ke1=npy.dot(P,load.static_matrix[:3,:])
                Ke2=npy.dot(L,npy.dot(P,load.static_matrix[:3,:]))+npy.dot(P,load.static_matrix[3:,:])
                Ke=npy.vstack([Ke1,Ke2])
                for indof,ndof in enumerate(self.sdof[load]):
                    K[ip*6:(ip+1)*6,ndof]=Ke[:,indof]     # minus because of sum is in LHS
                    
                    
            # knowns loads contribution to the RHS
            try:
                loads=loads_parts[part]
            except:
                loads=[]                
            for load in loads:
                P=geometry.Euler2TransferMatrix(*load.euler_angles)            
                u=load.position
                uprime=u
                L=geometry.CrossProductMatrix(uprime)
                F1=npy.dot(P,load.forces)
                F2=npy.dot(L,npy.dot(P,load.forces))+npy.dot(P,load.torques)
                Fe=npy.hstack([F1.reshape(3),F2.reshape(3)])
                F[ip*6:(ip+1)*6]-=Fe # minus because of sum is in LHS
        
        neq=6*lparts
        neq_linear=neq
        indices_r=list(range(neq))
        
        # behavior equations of linkages
        for linkage in self.linkages:
            neq_linkage=linkage.static_behavior_occurence_matrix.shape[0]
            if neq_linkage>0:
                # Adding to occurence matrix
                Me=npy.zeros((neq_linkage,self.n_sdof))
                for indof,ndof in enumerate(self.sdof[linkage]):
                    Me[:,ndof]+=linkage.static_behavior_occurence_matrix[:,indof]    
                M=npy.vstack([M,Me])

                # Adding linear equations to System matrix K
                neq_linear_linkage=linkage.static_behavior_linear_eq.shape[0]
                if neq_linear_linkage>0:
                    Ke=npy.zeros((neq_linear_linkage,self.n_sdof))
                    for indof,ndof in enumerate(self.sdof[linkage]):
                        Ke[neq_linear:neq_linear+6,ndof]+=linkage.static_behavior_linear_eq[:,indof]   
                    K=npy.vstack([K,Ke])
                    indices_r.extend(range(neq_linear,neq_linear+neq_linear_linkage))

                # Collecting non linear equations
#                print(v)
                if linkage.holonomic:
                    if linkage.static_require_kinematic:
                        # Relatives speed in this case
                        ls=self.LocalLinkageSpeeds(linkage)
                        w=ls[:3]
                        v=ls[3:]
                    else:
                        w,v=0,0
                    for i,fct in zip(linkage.static_behavior_nonlinear_eq_indices,linkage.static_behavior_nonlinear_eq):
                        nonlinear_eq[neq+i]=lambda x,v=v,w=w,fct=fct:fct(x,w,v)
                else:
                    if linkage.static_require_kinematic:
                        # Absolute speed in this case in local coordinate system
                        s=self.Speeds(linkage.position,self.ground,linkage.part1)
                        w=npy.dot(linkage.P.T,s[:3])
                        v=npy.dot(linkage.P.T,s[3:])
                    else:
                        w,v=0,0

                    for i,fct in zip(linkage.static_behavior_nonlinear_eq_indices,linkage.static_behavior_nonlinear_eq):
                        nonlinear_eq[neq+i]=lambda x,v=v,w=w,fct=fct:fct(x,w,v)
                        
                # Updating counters
                neq+=neq_linkage
                neq_linear+=neq_linear_linkage
                
                

        # behavior equations of unknowns loads
        for load in self.unknown_static_loads:
            neq_load=load.static_behavior_occurence_matrix.shape[0]
            if neq_load>0:
                # Adding to occurence matrix
                Me=npy.zeros((neq_load,self.n_sdof))
                for indof,ndof in enumerate(self.sdof[load]):
                    Me[:,ndof]+=load.static_behavior_occurence_matrix[:,indof]    
                M=npy.vstack([M,Me])

                # Adding linear equations to System matrix K
                neq_linear_load=load.static_behavior_linear_eq.shape[0]
                if neq_linear_load>0:
                    Ke=npy.zeros((neq_linear_load,self.n_sdof))
                    for indof,ndof in enumerate(self.sdof[load]):
                        Ke[neq_linear:neq_linear+6,ndof]+=load.static_behavior_linear_eq[:,indof]   
                    K=npy.vstack([K,Ke])
                    indices_r.extend(range(neq_linear,neq_linear+neq_linear_load))

                # Collecting non linear equations

                if load.static_require_kinematic:
                    # Absolute speed in this case in local coordinate system
                    s=self.Speeds(load.position,self.ground,load.part)
                    w=npy.dot(load.P.T,s[:3])
                    v=npy.dot(load.P.T,s[3:])
                else:
                    w,v=0,0
                for i,fct in zip(load.static_behavior_nonlinear_eq_indices,load.static_behavior_nonlinear_eq):
                    nonlinear_eq[neq+i]=lambda x,v=v,w=w,fct=fct:fct(x,w,v)
                        
                # Updating counters
                neq+=neq_load
                neq_linear+=neq_linear_load

        
        solvable,solvable_var,resolution_order=tools.EquationsSystemAnalysis(M,None)
#        print(resolution_order)
        if not solvable:
            raise ModelError('Overconstrained system')
        for eqs,variables in resolution_order:
#            print(eqs,variables)
            linear=True
            linear_eqs=[]
            for eq in eqs:
                try:
                    nonlinear_eq[eq]
                    linear=False
                except KeyError:
                    linear_eqs.append(eq)            
            if linear:
                eqs_r=npy.array([indices_r[eq] for eq in eqs])
                other_vars=npy.array([i for i in range(self.n_sdof) if i not in variables])
                Kr=K[eqs_r[:,None],npy.array(variables)]
                Fr=F[eqs_r]-npy.dot(K[eqs_r[:,None],other_vars],q[other_vars])

                q[variables]=linalg.solve(Kr,Fr)
            else:
                nl_eqs=[]
                other_vars=npy.array([i for i in range(self.n_sdof) if i not in variables])
                for eq in eqs:
                    try:
                        f1=nonlinear_eq[eq]
                        vars_func=[i for i in range(self.n_sdof) if M[eq,i]]
                        def f2(x,f1=f1,vars_func=vars_func,variables=variables,q=q):                    
                            x2=[]
                            for variable in vars_func:
                                try:
                                    x2.append(x[variables.index(variable)])
                                except ValueError:
                                    x2.append(q[variable])
                            return f1(x2)
                        nl_eqs.append(f2)
                    except KeyError:
                        # lambdification of linear equations
                        f2=lambda x,indices_r=indices_r,K=K,F=F,eq=eq,q=q,other_vars=other_vars:npy.dot(K[indices_r[eq],variables],x)-F[indices_r[eq]]+npy.dot(K[indices_r[eq],other_vars],q[other_vars])
                        nl_eqs.append(f2)
                f=lambda x:[fi(x) for fi in nl_eqs]
                xs=fsolve(f,npy.zeros(len(variables)),full_output=0)
                if npy.sum(npy.abs(f(xs)))>1e-4:
                    raise ModelError('No convergence of nonlinear phenomena solving'+str(npy.sum(npy.abs(f(xs)))))
                q[variables]=xs


        results={}
        for link,dofs in self.sdof.items():
            rlink={}
            for idof,dof in enumerate(dofs):
                rlink[idof]=q[dof]
            results[link]=rlink
            
        return results        
        
    def _KinematicAnalysis(self):
        # Kinematic setting
        self.n_kdof=0
        self.kdof={}        
        loops=nx.cycle_basis(self.holonomic_graph)
        ll=len(loops)
        ieq=6*ll
        neq=ieq+len(self.imposed_speeds)
        nhl=[]
        for linkage in self.linkages:
            try:
                # Holonomic linkage
                ndofl=linkage.kinematic_matrix.shape[1]
                for dofl in range(ndofl):
                    try:
                        self.kdof[linkage].append(self.n_kdof+dofl)
                    except KeyError:
                        self.kdof[linkage]=[self.n_kdof+dofl]                        
                self.n_kdof+=ndofl
            except AttributeError:
                # Non holonomic linkage
                nhl.append(linkage)
                neq+=len(linkage.kinematic_directions)
                        
        K=npy.zeros((neq,self.n_kdof))
        F=npy.zeros((neq,1))
        q=npy.zeros((self.n_kdof,1))
        M=npy.zeros((neq,self.n_kdof))
        for il,loop in enumerate(loops):
            for ilk,linkage in enumerate(loop):
                if not linkage.__class__.__name__=='Part':
                    try:
                        if loop[ilk+1]==linkage.part2:
                            side=1                            
                        else:
                            side=-1
                    except IndexError:
                        # linkage is last element of list
                        if loop[ilk-1]==linkage.part1:
                            side=1     
                        else:
                            side=-1
                    # Linkage
                    P=geometry.Euler2TransferMatrix(*linkage.euler_angles)            
                    u=linkage.position
                    uprime=u
                    L=geometry.CrossProductMatrix(uprime)
                    Me1=npy.abs(npy.dot(P,linkage.kinematic_matrix[:3,:]))>1e-10
                    Me2=npy.abs(npy.dot(L,npy.dot(P,linkage.kinematic_matrix[:3,:]))+npy.dot(P,linkage.kinematic_matrix[3:,:]))>1e-10
                    Me=npy.vstack([Me1,Me2])
                    for indof,ndof in enumerate(self.kdof[linkage]):
                        M[6*il:6*il+6,ndof]+=Me[:,indof]
                    

                    Ke1=npy.dot(P,linkage.kinematic_matrix[:3,:])
                    Ke2=npy.dot(L,npy.dot(P,linkage.kinematic_matrix[:3,:]))+npy.dot(P,linkage.kinematic_matrix[3:,:])
                    Ke=side*npy.vstack([Ke1,Ke2])
                    for indof,ndof in enumerate(self.kdof[linkage]):
                        K[6*il:6*il+6,ndof]+=Ke[:,indof]
        
        # Non holonomic equations
        for linkage in nhl:
            # Speed computation
            try:
                path=nx.shortest_path(self.holonomic_graph,linkage.part1,linkage.part2)
            except nx.NetworkXNoPath:
                raise ModelError('No path between {} and {} for linkage {} of type {}'.format(linkage.part1.name, linkage.part2.name, linkage.name, linkage.__class__.__name__))
            V=npy.zeros((3,self.n_kdof))
            for il, linkage2 in enumerate(path):
                if not linkage2.__class__.__name__=='Part':
                    try:
                        if path[il+1]==linkage2.part2:
                            side=1                            
                        else:
                            side=-1
                    except IndexError:
                        # linkage is last element of list
                        if path[il-1]==linkage2.part1:
                            side=1     
                        else:
                            side=-1
                    # It's really a linkage
                    P=geometry.Euler2TransferMatrix(*linkage2.euler_angles)            
                    u=linkage2.position
                    uprime=u-linkage.position
                    L=geometry.CrossProductMatrix(uprime)
                    Ve=npy.dot(L,npy.dot(P,linkage2.kinematic_matrix[:3,:]))+npy.dot(P,linkage2.kinematic_matrix[3:,:])
                    for indof,ndof in enumerate(self.kdof[linkage2]):
                        V[:,ndof]+=side*Ve[:,indof]
                        
                    
            P=geometry.Euler2TransferMatrix(*linkage.euler_angles)            
            for direction in linkage.kinematic_directions:
                K[ieq,:]=npy.dot(npy.dot(P,direction),V)
                ieq+=1
        
        # Imposed speeds equations
        for linkage,index,speed in self.imposed_speeds:
            K[ieq,self.kdof[linkage][index]]=1
            F[ieq,0]=speed
            ieq+=1
        
        # deducing M from K for last lines
        M[6*ll:,:]=npy.abs(K[6*ll:,:])>1e-10
        
        solvable,solvable_var,resolution_order=tools.EquationsSystemAnalysis(M,None)

        if solvable:
            for eqs,variables in resolution_order:
                eqs=npy.array(eqs)
                other_vars=npy.array([i for i in range(self.n_kdof) if i not in variables])
                Kr=K[eqs[:,None],npy.array(variables)]
                Fr=F[eqs,:]-npy.dot(K[eqs[:,None],other_vars],q[other_vars,:])
                q[variables,:]=linalg.solve(Kr,Fr)
            results={}
            for link,dofs in self.kdof.items():
                rlink={}
                for idof,dof in enumerate(dofs):
                    rlink[idof]=q[dof,0]
                results[link]=rlink
            return results,q
        else:
            raise ModelError
    
    def GlobalSankey(self):
        from matplotlib.sankey import Sankey
        flows=[]
        orientations=[]
        labels=[]
        
        
        for load in self.known_static_loads+self.unknown_static_loads:
            pl=self.LoadPower(load)
            print(pl)
            flows.append(pl)
            labels.append(load.name)
            if (load.__class__.__name__=='SimpleUnknownLoad')|(load.__class__.__name__=='KnownLoad'):
                orientations.append(0)
            else:
                orientations.append(-1)
                
        for linkage in self.linkages:
            pl=self.LinkagePowerLosses(linkage)
#            print(pl)
            if pl!=0.:
                flows.append(-pl)
                orientations.append(-1)
                labels.append(linkage.name)
#            pl.append(0.5*l)
        l=max([abs(f) for f in flows])
#        pl=[0.1*l]*len(flows)

        sankey = Sankey(unit='W',scale=1/l)
    
        sankey.add(flows=flows,
           orientations=orientations,labels=labels,)
        
        sankey.finish()
        
    def VMPlot(self,u=1):
        import volmdlr as vm
        points=[]
        for linkage in self.linkages:
            points.append(vm.Point3D(linkage.position))
            points.append(vm.Line3D(vm.Point3D(linkage.position),vm.Point3D(linkage.position+u*linkage.P[:,0])))
        mdl=vm.VolumeModel(points)
        mdl.MPLPlot()
        
    def SceneCaracteristics(self):
        min_vect=self.linkages[0].position.copy()
        max_vect=self.linkages[0].position.copy()
        center=self.linkages[0].position.copy()
        n=1
        for linkage in self.linkages[1:]+self.known_static_loads+self.unknown_static_loads:
#            print(linkage)
            for i,(xmin,xmax,xi) in enumerate(zip(min_vect,max_vect,linkage.position)):
#                print(i,xmin,xmax,xi)
                if xi<xmin:
                    min_vect[i]=xi
#                    print('min',min_vect)
                if xi>xmax:
                    max_vect[i]=xi
#                    print('max',max_vect)
#            print(linkage,linkage.position)
            center+=linkage.position
            n+=1
        
        center=center/n
#        print(min_vect,max_vect)
        max_length=linalg.norm(min_vect-max_vect)
#        print(center,max_length)
        return center,max_length
    
            
    def BabylonScript(self,forces=True):

        env = Environment(loader=PackageLoader('genmechanics', 'templates'),
                          autoescape=select_autoescape(['html', 'xml']))
        
        template = env.get_template('babylon.html')
        
        center,length=self.SceneCaracteristics()
        
        if forces:
            max_force=0.
            max_torque=0
            for linkage in self.linkages:
                f=self.GlobalLinkageForces(linkage,0)[0:3]
                t=self.GlobalLinkageForces(linkage,0)[3:]
                max_force=max(max_force,max(f))
                max_torque=max(max_torque,max(t))
            
            for load in self.known_static_loads+self.unknown_static_loads:
                f=self.GlobalLoadForces(load)[0:3]
                t=self.GlobalLoadForces(load)[3:]
                max_force=max(max_force,max(f))
                max_torque=max(max_torque,max(t))
                
            print(max_force,max_torque)
        
                
        linkages_strings=[]
        for linkage in self.linkages:
            try:
                if forces:
#                    print([f/max_force/length/4 for f in self.GlobalLoadForces(load)[0:3]])
                    linkages_strings.append(linkage.Babylon(length,
                                                            [f/max_force*length/4 for f in self.GlobalLinkageForces(linkage,0)[0:3]],
                                                            [f/max_torque*length/4 for f in self.GlobalLinkageForces(linkage,0)[3:]]))
                else:
                    linkages_strings.append(linkage.Babylon(length,None,None))

            except AttributeError:
                print('error')
                pass
        
        
            
            
        return template.render(name=self.name,center=tuple(center),length=length,
                               linkages_strings=linkages_strings)
    
    def BabylonShow(self,page='gm_babylonjs',forces=True):
        page+='.html'
#        print(self.BabylonScript())
        with open(page,'w') as file:
            file.write(self.BabylonScript())
        
        webbrowser.open('file://' + os.path.realpath(page))
    
#    def BabylonJS(self):
#        center,length=self.SceneCaracteristics()
#        s="""
#        <!doctype html>
#<html>
#<head>
#   <meta charset="utf-8">
#   <title>Babylon - Basic scene</title>
#   <style>
#      html, body {
#         overflow: hidden;
#         width: 100%;
#         height: 100%;
#         margin: 0;
#         padding: 0;
#      }
#      #renderCanvas {
#         width: 100%;
#         height: 100%;
#         touch-action: none;
#      }
#   </style>
#   <script src="https://cdnjs.cloudflare.com/ajax/libs/babylonjs/2.5.0/babylon.js"></script>
#   <script src="https://cdnjs.cloudflare.com/ajax/libs/handjs/1.3.11/hand.js"></script>
#   <script src="https://cdnjs.cloudflare.com/ajax/libs/cannon.js/0.6.2/cannon.min.js"></script> <!-- optional physics engine -->
#</head>
#<body>
#   <canvas id="renderCanvas"></canvas>
#   <script type="text/javascript">
#      // Get the canvas element from our HTML below
#      var canvas = document.querySelector("#renderCanvas");
#      // Load the BABYLON 3D engine
#      var engine = new BABYLON.Engine(canvas, true);
#      // -------------------------------------------------------------
#      // Here begins a function that we will 'call' just after it's built
#      var createScene = function () {
#         // Now create a basic Babylon Scene object
#         var scene = new BABYLON.Scene(engine);
#         // This creates and positions a free camera
#         //var plane = BABYLON.Mesh.CreatePlane("plane", 10.0, scene);
#         """
#         
#         
##        s+='var camera = new BABYLON.FreeCamera("camera1", new BABYLON.Vector3({}, {}, {}), scene);'.format(*center)
#        s+='var camera = new BABYLON.ArcRotateCamera("ArcRotateCamera", 0., 0., {}, new BABYLON.Vector3({}, {}, {}), scene);\n'.format(4*length,*center)
##        s+='var camera = new BABYLON.ArcRotateCamera("ArcRotateCamera", 0., 0., 15, new BABYLON.Vector3(0,0,0), scene);'
#        s+='camera.panningSensibility={};\n'.format(30)
#        s+='camera.pinchPrecision={};\n'.format(20)
#        s+='camera.wheelPrecision={};\n'.format(20)
#        s+="""
#        camera.attachControl(canvas, false);
#         // This creates a light, aiming 0,1,0 - to the sky.
#         var light = new BABYLON.HemisphericLight("light1", new BABYLON.Vector3(0, 1, 0), scene);
#         // Dim the light a small amount
#         light.intensity = .5;
#         """
#         
##         // Let's try our built-in 'sphere' shape. Params: name, subdivisions, size, scene
##         var sphere = BABYLON.Mesh.CreateSphere("sphere1", 16, 2, scene);
##         // Move the sphere upward 1/2 its height
##         sphere.position.y = 1;
##         // Let's try our built-in 'ground' shape. Params: name, width, depth, subdivisions, scene
##         var ground = BABYLON.Mesh.CreateGround("ground1", 6, 6, 2, scene);
#
#        for linkage in self.linkages:
#            s+='var sphere = BABYLON.Mesh.CreateSphere("sphere1", 15., {}, scene);\n'.format(length/30)
#            s+="sphere.position=new BABYLON.Vector3({},{},{});\n".format(*linkage.position)    
##            s+="var lines = BABYLON.Mesh.CreateLines("lines", [    new BABYLON.Vector3(-10, 0, 0)"
#        s+="""
#         // Leave this function
#         return scene;
#      }; // End of createScene function
#      // -------------------------------------------------------------
#      // Now, call the createScene function that you just finished creating
#      var scene = createScene();
#      // Register a render loop to repeatedly render the scene
#      engine.runRenderLoop(function () {
#         scene.render();
#      });
#      // Watch for browser/canvas resize events
#      window.addEventListener("resize", function () {
#         engine.resize();
#      });
#   </script>
#</body>
#</html>
#
#        """
#        with open('gm_babylonjs.html','w') as file:
#            file.write(s)
#        
#        webbrowser.open('file://' + os.path.realpath('gm_babylonjs.html'))
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 24 10:50:15 2016

@author: steven
"""


import genmechanics
import genmechanics.linkages as linkages
import numpy as npy

L=0.2
e1=0.05
C=300
r1=0.8
W=300

Ca=0.0008
Cr=0.0006

ground=genmechanics.Part('ground')
shaft1=genmechanics.Part('shaft1')
shaft2=genmechanics.Part('shaft2')

p1a=npy.array([0,0,0])
p1b=npy.array([L,0,0])
p2a=npy.array([0,e1,0])
p2b=npy.array([L,e1,0])

pgs1=0.5*(p1a+p1b)*r1+(1-r1)*0.5*(p2a+p2b)

dgs1=npy.cross(p1b-p1a,p2a-p1a)
egs1=genmechanics.geometry.Direction2Euler(*dgs1)

bearing1=linkages.FrictionlessRevoluteLinkage(ground,shaft1,p1a,[0,0,0],'bearing1a')
#bearing1=linkages.LinearAnnularLinkage(ground,shaft1,p1b,[0,0,0],Cr,'bearing1b')
bearing2=linkages.FrictionlessRevoluteLinkage(ground,shaft2,p2a,[0,0,0],'bearing2a')
#bearing2b=linkages.LinearAnnularLinkage(ground,shaft2,p2b,[0,0,0],Cr,'bearing2b')

gearset12=linkages.FrictionlessGearSetLinkage(shaft1,shaft2,pgs1,egs1,'Gear set 2')

load1=genmechanics.KnownMechanicalLoad(shaft1,[-L/4,0,0],[0,0,0],[0,0,0],[C,0,0],'input torque')
load2=genmechanics.UnknownMechanicalLoad(shaft2,[L/2,0,0],[0,0,0],[],[0],'output torque')
imposed_speeds=[(bearing1,0,W)]

mech=genmechanics.Mechanism([bearing1,bearing2,gearset12],ground,imposed_speeds,[load1],[load2])


for l,lv in mech.static_results.items():
    for d,v in lv.items():
        print(l.name,d,v)
print('=================')
for l,lv in mech.kinematic_results.items():
    for d,v in lv.items():
        print(l.name,d,v)
        
        
print('wth: ',(1-r1)/r1*W)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Dec 25 09:39:37 2016

@author: steven
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 24 10:50:15 2016

@author: steven
"""


import genmechanics
import genmechanics.linkages as linkages
import genmechanics.loads as loads
import numpy as npy

L=2.2
C=300
Fa=5000
Fr=2500
W=300
Cwb=0.00001# Speed coeff for bearings

Ca=0.0008
Cr=0.0006

ground=genmechanics.Part('ground')
shaft1=genmechanics.Part('shaft1')

p1=npy.array([0,0,0])
p2=npy.array([L,0,0])

#bearing1=linkages.BallLinkage(ground,shaft1,p1,[0,0,0],Cr,Ca,Cwb,'bearing1a')
bearing1=linkages.FrictionlessBallLinkage(ground,shaft1,p1,[0,0,0],'bearing1a')
#bearing1=linkages.LinearAnnularLinkage(ground,shaft1,p1b,[0,0,0],Cr,'bearing1b')
bearing2=linkages.FrictionlessLinearAnnularLinkage(ground,shaft1,p2,[0,0,0],'bearing2a')
#bearing2=linkages.BallLinkage(ground,shaft1,p2,[0,0,0],Cr,Ca,Cwb,'bearing2a')
#bearing2b=linkages.LinearAnnularLinkage(ground,shaft2,p2b,[0,0,0],Cr,'bearing2b')

load1=loads.KnownLoad(shaft1,[L/2,10,0],[0,0,0],[Fa,Fr,Fr/2.],[10,100,0],'input torque and loads')
load3=loads.KnownLoad(shaft1,[L/2,0,0],[0,0,0],[Fa,Fr,Fr/2.],[0,0,0],'input torque and loads')
load2=loads.SimpleUnknownLoad(shaft1,[L/2,0,0],[0,0,0],[],[0],'output torque')
imposed_speeds=[(bearing1,0,W)]

mech=genmechanics.Mechanism([bearing1,bearing2],ground,imposed_speeds,[load1,load3],[load2])
        
##K=mech.static_results['K']
for l,lv in mech.static_results.items():
    for d,v in lv.items():
        print(l.name,d,v)
#print('=================')
for l,lv in mech.kinematic_results.items():
    for d,v in lv.items():
        print(l.name,d,v)
        
        


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 16 14:21:07 2016

@author: steven
"""

import genmechanics
import genmechanics.linkages as linkages
import genmechanics.loads as loads

import numpy as npy
import scipy.linalg as linalg

Lb=0.2# bearing width
Lgs=0.045# Gearset width


C=300
w=300
r1=0.048
r2=0.047
y2=0.078
z2=-0.026
y3=0.203
z3=0


Ca=0.0008
Cr=0.0006
Cf=0.01
Cwb=0.00001# Speed coeff for bearings
Cvgs=0.0001# Speed coeff for gear sets

alpha_gs1=20/360*2*3.1415
beta_gs1=25/360*2*3.1415
alpha_gs2=20/360*2*3.1415
beta_gs2=-25/360*2*3.1415


ground=genmechanics.Part('ground')
shaft1=genmechanics.Part('shaft1')
shaft2=genmechanics.Part('shaft2')
shaft3=genmechanics.Part('shaft3')

p1a=npy.array([0.,0.,0.])
p1b=npy.array([Lb+Lgs,0.,0.])
p2a=npy.array([0,y2,z2])
p2b=npy.array([Lb+2*Lgs,y2,z2])
p3a=npy.array([0,y3,z3])
p3b=npy.array([Lb+2*Lgs,y3,z3])

pgs11=npy.array([0.5*(Lgs+Lb),0,0])
pgs12=npy.array([0.5*(Lgs+Lb),y2,z2])
pgs22=npy.array([0.5*(3*Lgs+Lb),y2,z2])
pgs23=npy.array([0.5*(3*Lgs+Lb),y3,z3])

pgs1=pgs11+(pgs12-pgs11)/linalg.norm(pgs12-pgs11)*r1
pgs2=pgs22+(pgs23-pgs22)/linalg.norm(pgs23-pgs22)*r2

dir_axis=npy.array([1,0,0])

dgs1=npy.cross(p1b-p1a,p2a-p1a)
egs1=genmechanics.geometry.Direction2Euler(dgs1,dir_axis)
dgs2=npy.cross(p3b-p2a,p3a-p2a)
egs2=genmechanics.geometry.Direction2Euler(dgs2,dir_axis)

bearing1a=linkages.BallLinkage(ground,shaft1,p1a,[0,0,0],Ca,Cr,Cwb,'bearing1a')
bearing1b=linkages.LinearAnnularLinkage(ground,shaft1,p1b,[0,0,0],Cr,Cwb,'bearing1b')
bearing2a=linkages.BallLinkage(ground,shaft2,p2a,[0,0,0],Ca,Cr,Cwb,'bearing2a')
bearing2b=linkages.LinearAnnularLinkage(ground,shaft2,p2b,[0,0,0],Cr,Cwb,'bearing2b')
bearing3a=linkages.BallLinkage(ground,shaft3,p3a,[0,0,0],Ca,Cr,Cwb,'bearing3a')
bearing3b=linkages.LinearAnnularLinkage(ground,shaft3,p3b,[0,0,0],Cr,Cwb,'bearing3b')


gearset12=linkages.GearSetLinkage(shaft1,shaft2,pgs1,egs1,alpha_gs1,beta_gs1,Cf,Cvgs,'Gear set 1')
gearset23=linkages.GearSetLinkage(shaft2,shaft3,pgs2,egs2,alpha_gs2,beta_gs2,Cf,Cvgs,'Gear set 2')

load1=loads.KnownLoad(shaft1,[-Lb/2,0,0],[0,0,0],[0,0,0],[C,0,0],'input torque')
load2=loads.SimpleUnknownLoad(shaft3,[2*(Lgs+Lb),y2,z2],[0,0,0],[],[0],'output torque')

imposed_speeds=[(bearing1a,0,w)]

mech=genmechanics.Mechanism([bearing1a,bearing1b,bearing2a,bearing2b,bearing3a,bearing3b,gearset12,gearset23],ground,imposed_speeds,[load1],[load2])


for l,lv in mech.static_results.items():
    for d,v in lv.items():
        print(l.name,d,v)

#print('Cth: ',r1/(1-r1)*r2/(1-r2)*C)# False now

for l,lv in mech.kinematic_results.items():
    for d,v in lv.items():
        print(l.name,d,v)

#print('wth: ',(1-r1)/r1*(1-r2)/r2*w) # False now

mech.GlobalSankey()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Springs models

"""
import numpy as npy
import scipy.linalg as linalg
import matplotlib.pyplot as plt
import networkx as nx
import random

from matplotlib.patches import Arrow, Polygon

class Body:
    """
    :param y_position: The y position in plots of the body. 
    It has no effect on the computation
    """
    def __init__(self, initial_position, y_position=0., name=''):
        self.initial_position = initial_position
        self.y_position = y_position
        self.name = name
        
class Load:
    def __init__(self, body, value):
        self.body = body
        self.value = value
        
    def Plot(self, bodies_position=None, ax=None, width=0.008, intensity_factor=1e-4):
        if bodies_position is None:
            bodies_position = {self.body: self.body.initial_position}
        x = bodies_position[self.body]
        y = self.body.y_position
        ax.add_patch(Arrow(x, y,
                     self.value*intensity_factor, 0, width))
        ax.text(x+0.5*self.value*intensity_factor, y+0.2*width,
                '{} N'.format(round(self.value)),
                color='k', horizontalalignment='center', verticalalignment='bottom')
        
class ImposedDisplacement:
    def __init__(self, body, value):
        self.body = body
        self.value = value
        self.neqs = 1
        
    def Plot(self, bodies_positions=None, strains=None, ax=None, 
             width=0.01, color='k', intensity_factor=1e-6):
        if bodies_positions is None:
            bodies_positions = {self.body: self.body}
        
        x0 = self.body.initial_position
        x1 = bodies_positions[self.body]
        y = self.body.y_position

        h = 0.5*width
        ax.add_patch(Polygon([(x1, y), (x1-h, y-h), (x1+h, y-h)], True, color='grey'))
        if self.value != 0:
            ax.add_patch(Arrow(x0, y, self.value, 0, width = 0.5*h, color='g'))        
            ax.text(x0 + 0.5*self.value, y+0.2*h,
                    '{} mm'.format(round(1000*self.value)),
                    horizontalalignment='center', verticalalignment='bottom')
            
            
class Linkage:
    def __init__(self, body1, body2, n_constraints, name=''):
        self.body1 = body1
        self.body2 = body2
        self.n_constraints = n_constraints
        self.name = name 
        
class Spring(Linkage):
    def __init__(self, body1, body2, stiffness, free_length, name=''):
        Linkage.__init__(self, body1, body2, 0, name) 
        self.stiffness = stiffness
        self.free_length = free_length
        
    def StiffnessMatrices(self):
        return (self.stiffness * npy.array([[1,-1], [-1, 1]]),
                self.stiffness * self.free_length * npy.array([-1, 1]))
        
    def ConstraintsMatrices(self):
                return None, None
        
    def Plot(self, bodies_positions=None, strains=None, ax=None,
             linkages_width=0.01, color='k', intensity_factor=1e-6):
        if bodies_positions is None:
            bodies_positions = {self.body1: self.body1.initial_position,
                                self.body2: self.body2.initial_position}
            
        x1 = bodies_positions[self.body1]
        x2 = bodies_positions[self.body2]
        y1 = self.body1.y_position
        y2 = self.body2.y_position
        ax.plot([x1, x2], [y1, y2], linestyle='--', color=color)
        
        xm = 0.5*(x1 + x2)
        ym = 0.5*(y1 + y2)
        
        ax.text(xm, ym-0.5*linkages_width,
                '{} #{}'.format(self.__class__.__name__, str(self.__hash__())[-5:]),
                horizontalalignment='center')

        
    def Strains(self, positions):
        return self.stiffness*(self.free_length - (positions[1] - positions[0]))
        

class CompressionSpring(Spring):
    def __init__(self, body1, body2, stiffness, free_length, name=''):
        Linkage.__init__(self, body1, body2, 0, name) 
        self.stiffness = stiffness
        self.free_length = free_length
        
    def StiffnessMatrices(self):
        return (self.stiffness * npy.array([[1,-1], [-1, 1]]),
                self.stiffness * self.free_length * npy.array([-1, 1]))
        
    def ConstraintsMatrices(self):
        return None, None
        
    def ActivationCondition(self, positions, reactions):
        return (positions[1] - positions[0]) < self.free_length + 1e-10

    def DesactivationCondition(self, positions, reactions):
        return (positions[1] - positions[0]) > self.free_length + 1e-10

class UnilateralContact(Linkage):
    def __init__(self, body_left, body_right, contact_distance, name=''):
        Linkage.__init__(self, body_left, body_right, 1, name) 
        self.contact_distance = contact_distance
    
    def StiffnessMatrices(self):
        return None, None
    
    def ConstraintsMatrices(self):
        return npy.array([[1.,-1.]]), npy.array([0.])
        
    def ActivationCondition(self, positions, reactions):
#        print('act', self.name, p, (p[1] - p[0]) < self.contact_distance)
        return (positions[1] - positions[0]) <= self.contact_distance-1e-8

    def DesactivationCondition(self, positions, reactions):        
#        print('dis', self.name, r, r < -1e-10)
        return reactions[0] < 1e-10
        
    def Plot(self, bodies_positions=None, strains=None, ax=None,
             linkages_width=0.01, color='k', intensity_factor = 1e-6):
        x1 = bodies_positions[self.body1]
        x2 = bodies_positions[self.body2]
        y1 = self.body1.y_position
        y2 = self.body2.y_position
        ym = 0.5*(y1+y2)
#        h = max(0.6*abs(y1-y2), linkages_width)
#        l = ((x2-x1)**2+(y2-y1)**2)**0.5
        h = 0.5*linkages_width
        
        if self.contact_distance > 0:
            xplate1 = x1 + 0.5*(self.contact_distance)
            xplate2 = x2 - 0.5*(self.contact_distance)
            xt1 = [x1, xplate1, xplate1, xplate1]
            yt1 = [y1, ym, ym+h, ym-h]
            xt2 = [x2, xplate2, xplate2, xplate2]
            yt2 = [y2, ym, ym+h, ym-h]
        else:
            xplate1 = x1 + 0.5*(self.contact_distance)
            xplate2 = x2 - 0.5*(self.contact_distance)
            xturn1 = x1 + 0.55*(self.contact_distance)
            xturn2 = x2 - 0.55*(self.contact_distance)
            xt1 = [x1, xturn1, xturn1, xplate1, xplate1, xplate1]
            yt1 = [y1, y1, ym, ym, ym+h, ym-h]
            xt2 = [x2, xturn2, xturn2, xplate2, xplate2, xplate2]
            yt2 = [y2, y2, ym, ym, ym+h, ym-h]
        
            
        ax.plot(xt1, yt1, color=color)
        ax.plot(xt2, yt2, color=color)
        ax.text(0.5*(x1+x2), ym-0.5*linkages_width,
                'Unilateral contact #{}'.format(str(self.__hash__())[-5:]),
                horizontalalignment='center')
        
        if strains is not None:
            if self in strains:
                strain = strains[self][0]
           
            
                if strain < 1e-6:
                    ax.add_patch(Arrow(xplate1, ym,
                     -strain*intensity_factor, 0, linkages_width, color='r'))
                    ax.add_patch(Arrow(xplate2, ym,
                     strain*intensity_factor, 0, linkages_width, color='r'))
                elif strain > -1e-6:
                    ax.add_patch(Arrow(xplate1 - strain*intensity_factor, ym,
                     strain*intensity_factor, 0, linkages_width, color='r'))
                    ax.add_patch(Arrow(xplate2 + strain*intensity_factor, ym,
                     -strain*intensity_factor, 0, linkages_width, color='r'))
                
                if abs(strain) > 1e-6:
                    ax.text(xplate1, ym+0.5*linkages_width,
                            '{} N'.format(round(strain, 3)), color='k',
                            horizontalalignment='center')
        
class ModelSolvingError(Exception):
    def __init__(self, strain_violation):

        # Call the base class constructor with the parameters it needs
        super().__init__('Residue of least squares to high')

#        self.displacement_violation = displacement_violation
        self.strain_violation = strain_violation
        
class ModelConvergenceError(Exception):
    def __init__(self, iters):

        # Call the base class constructor with the parameters it needs
        super().__init__('Number of iteration exceeded: {}'.format(iters))

#        self.displacement_violation = displacement_violation
        self.iters = iters
    
class UnidimensionalModel:
    """
    
    """
    def __init__(self, bodies, linear_linkages, nonlinear_linkages,
                 loads, imposed_displacements):
        self.bodies = bodies
        self.linear_linkages = linear_linkages
        self.nonlinear_linkages = nonlinear_linkages
        self.loads = loads
        self.imposed_displacements = imposed_displacements
#        self.nonlinear_imposed_displacements = nonlinear_imposed_displacements
        
        
    def Graph(self, activated_nonlinar_linkages=None):
        G = nx.Graph()
        G.add_nodes_from(self.bodies)
        if activated_nonlinar_linkages is None:
            activated_nonlinar_linkages = self.nonlinear_linkages[:]
        for linkage in self.linear_linkages:
            G.add_edge(linkage.body1, linkage.body2, linear=True)
        for linkage in activated_nonlinar_linkages:
            G.add_edge(linkage.body1, linkage.body2, linear=False)
        return G
    
    def PlotGraph(self, activated_nonlinar_linkages=None):
        if activated_nonlinar_linkages is None:
            activated_nonlinar_linkages = self.nonlinear_linkages
        plt.figure()
        G = self.Graph(activated_nonlinar_linkages)
        pos={}
        labels = {}
        for body in self.bodies:
            labels[body] = body.name
            pos[body] = (body.initial_position, body.y_position)
            
        linear_edges = []
        nonlinear_edges = []
        for e in G.edges():
            if G.edges()[e]['linear']:
                linear_edges.append(e)
            else:
                nonlinear_edges.append(e)
                
        nx.draw_networkx_nodes(G, pos)
        nx.draw_networkx_edges(G, pos, linear_edges)
        nx.draw_networkx_edges(G, pos, nonlinear_edges, style='--')

        nx.draw_networkx_labels(G, pos, labels)
        
    def IsModelValid(self, activated_nonlinar_linkages):
        """
        Check if each Load is connected to at least a force exit (imposed displacement)
        """
        output_bodies = [i_dis.body for i_dis in self.imposed_displacements]
        input_bodies = [l.body for l in self.loads]
#        print('i', input_bodies)
#        print('o', output_bodies)
       
        G = self.Graph(activated_nonlinar_linkages)
        for ib in input_bodies:
            has_exit = False
            for ob in output_bodies:
                if nx.has_path(G, ib, ob):
                    has_exit = True
#                    print('exit', ib.name, ob.name)
                    break
            if not has_exit:
                return False
        return True
            
        
    def Settings(self, activated_nonlinear_linkages):
        dof = {}
        for ibody, body in enumerate(self.bodies):
            dof[body] = ibody
        ndof = len(dof)
        
        n_constraints = 0
        for linkage in activated_nonlinear_linkages:
            n_constraints += linkage.n_constraints
        for imposed_displacement in self.imposed_displacements:
            n_constraints += imposed_displacement.neqs
        
        return dof, ndof, n_constraints
        
    def LinearSolve(self, activated_nonlinear_linkages=[], strains_tol=1.):
        
        dof, ndof, n_constraints = self.Settings(activated_nonlinear_linkages)
        K = npy.zeros((ndof, ndof))
        F = npy.zeros(ndof)
        D = npy.zeros((n_constraints, ndof))
        Fd = npy.zeros(n_constraints)
#        print(D)
        
        for linkage in self.linear_linkages+activated_nonlinear_linkages:
            dof_linkage1 = dof[linkage.body1]
            dof_linkage2 = dof[linkage.body2]
            Ke, Fe,  = linkage.StiffnessMatrices()
#            print(dof_linkage1, Ke.shape)
            if Ke is not None:
                K[dof_linkage1, dof_linkage1] += Ke[0, 0]
                K[dof_linkage1, dof_linkage2] += Ke[0, 1]
                K[dof_linkage2, dof_linkage1] += Ke[1, 0]
                K[dof_linkage2, dof_linkage2] += Ke[1, 1]

            if Fe is not None:            
                F[dof_linkage1] += Fe[0]
                F[dof_linkage2] += Fe[1]
            
        for load in self.loads:
            F[dof[load.body]] += load.value
        

        i_dof_constraint = ndof
        ic = 0
        for imposed_displacement in self.imposed_displacements:
            D[ic, dof[imposed_displacement.body]] = 1
            Fd[ic] = imposed_displacement.value
            dof[imposed_displacement] = i_dof_constraint
            i_dof_constraint += 1
            ic += 1
        
        for linkage in self.linear_linkages+activated_nonlinear_linkages:
            if linkage.n_constraints > 0:
                Ce, Fce = linkage.ConstraintsMatrices()
                dof_linkage1 = dof[linkage.body1]
                dof_linkage2 = dof[linkage.body2]
                for j in range(linkage.n_constraints):
                    dof[linkage] = i_dof_constraint
                    i_dof_constraint += 1
                    D[ic, dof_linkage1] += Ce[j, 0]
                    D[ic, dof_linkage2] += Ce[j, 1]
                    Fd[ic] += Fce[j]
                    ic += 1
        
        A = npy.vstack((npy.hstack((K, D.T)), 
                        npy.hstack((D, npy.zeros((n_constraints, n_constraints))))))
#        print(K)
#        print(D)
#        print(K.shape, A.shape, linalg.det(A))
#        return A
        b = npy.hstack((F, Fd))
#        print(A, b)
        if abs(linalg.det(A)) < 1e-10:
            res = linalg.lstsq(A, b)
#            print(res)
            x = res[0]
#            print('res', )
            residue = npy.dot(A,x)-b
#            displ_violation = []
            strains_violation = []
            for ires, res in enumerate(npy.abs(residue)):
#                print(ires, res)
                if ires < ndof:
                    if res > strains_tol:
                        strains_violation.append((ires, res))
#                else:
#                    if res > strains_tol:
#                        strains_violation.append((ires, res))
#            print(strains_violation)
            if len(strains_violation)> 0:
                raise ModelSolvingError(strains_violation)
        
        else:
            x = linalg.solve(A, b)
#        print(x)
        displacements = {}
        strains = {}
        for element, dof_element in dof.items():
#            print(element, dof_element)
            if dof_element < ndof:
                displacements[element] = x[dof_element]
            else:
                if element in strains:                    
                    strains[element].append(x[dof_element])
                else:
                    strains[element] = [x[dof_element]]

        return UnidimensionalModelResults(self, activated_nonlinear_linkages, displacements, strains)
    
    def Solve(self, max_iters=100):
        iters = 0
        activated_nonlinear_linkages = []
        new_activated_nonlinear_linkages = self.nonlinear_linkages[:]

        while (set(new_activated_nonlinear_linkages) != set(activated_nonlinear_linkages)) and iters < max_iters:

            activated_nonlinear_linkages = new_activated_nonlinear_linkages[:]
            additional_linkages = []
            while not self.IsModelValid(activated_nonlinear_linkages+additional_linkages):
                
                disactivated_linkages = [l for l in self.nonlinear_linkages if not l in activated_nonlinear_linkages]
                nla = random.choice(range(1, len(disactivated_linkages)))
                additional_linkages = list(random.sample(disactivated_linkages, nla))
            activated_nonlinear_linkages.extend(additional_linkages)
            
            try:
                result = self.LinearSolve(activated_nonlinear_linkages)
            except ModelSolvingError as e:
                raise RuntimeError
            
            new_activated_nonlinear_linkages = []
            disactivated_nonlinear_linkages = []
            for linkage in self.nonlinear_linkages:
                positions = (result.positions[linkage.body1], result.positions[linkage.body2])
                if linkage in result.strains:
                    reactions = result.strains[linkage]
                else:
                    reactions = None
                if linkage in activated_nonlinear_linkages:
                    # Checking if linkage should be desactivated
                    if not linkage.DesactivationCondition(positions, reactions):
                        new_activated_nonlinear_linkages.append(linkage)
                    else:
                        disactivated_nonlinear_linkages.append(linkage)
                else:
                    
                    if linkage.ActivationCondition(positions, reactions):
                        new_activated_nonlinear_linkages.append(linkage)
            iters += 1

        if max_iters == iters:
#            self.Plot()
#            self.PlotGraph()
#            print(self.linear_linkages, self.nonlinear_linkages)
#            print(self.imposed_displacements)
            raise ModelConvergenceError(iters)
            
        result = self.LinearSolve(activated_nonlinear_linkages)
        return result
        
            
    
    def Plot(self, activated_nonlinear_linkages = None, bodies_positions=None,
             strains={}, ax=None, color='k', linkages_width=0.006, intensity_factor=1e-6):
        
        if activated_nonlinear_linkages is None:
            activated_nonlinear_linkages = self.nonlinear_linkages
        
        if bodies_positions is None:
            bodies_positions = {b:b.initial_position for b in self.bodies}
            
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
            
        for body in self.bodies:
            ax.plot([bodies_positions[body]], [body.y_position], 'o', color=color)
            
        for linear_linkages in self.linear_linkages+activated_nonlinear_linkages:
            linear_linkages.Plot(bodies_positions, strains=strains, ax=ax, color=color,
                                 linkages_width=linkages_width, intensity_factor=intensity_factor)
            
        for load in self.loads:
            load.Plot(bodies_positions, ax=ax, width=linkages_width,
                      intensity_factor=intensity_factor)
            
        for imposed_displacement in self.imposed_displacements:
            imposed_displacement.Plot(bodies_positions, ax=ax, width=linkages_width,
                      intensity_factor=intensity_factor)
            
            
        ax.autoscale()
        return fig, ax
        
class UnidimensionalModelResults:
    def __init__(self, model, activated_nonlinear_linkages, displacements, strains):
        self.model = model
        self.activated_nonlinear_linkages = activated_nonlinear_linkages
        self.displacements = displacements
        self.strains = strains
        
        self.positions = {b:b.initial_position+displacements[b] for b in self.model.bodies}
        
    def Plot(self, intensity_factor=1e-6):
        fig, ax = plt.subplots()
        self.model.Plot(self.activated_nonlinear_linkages,
                        bodies_positions=self.positions,
                        strains=self.strains,
                        ax=ax, intensity_factor=intensity_factor)
        
        
    


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 16 14:21:07 2016

@author: steven
"""

import genmechanics
import genmechanics.linkages as linkages
import genmechanics.loads as loads
import numpy as npy
import scipy.linalg as linalg

L=0.2
e1=0.05
e2=0.07
C=300
w=300
r1=0.3

Ca=0.0005
Cr=0.0001
Cf=0.01
Cwb=0.00001# Speed coeff for bearings
Cvgs=0.0001# Speed coeff for gear sets

alpha_gs1=18/360*2*3.1415
beta_gs1=20/360*2*3.1415

ground=genmechanics.Part('ground')
shaft1=genmechanics.Part('shaft1')
shaft2=genmechanics.Part('shaft2')

p1a=npy.array([0,0,0])
p1b=npy.array([L,0,0])
p2a=npy.array([0,e1,0])
p2b=npy.array([L,e1,0])


pgs1=0.5*(p1a+p1b)*r1+(1-r1)*0.5*(p2a+p2b)
print(pgs1)


bearing1a=linkages.BallLinkage(ground,shaft1,p1a,[0,0,0],Ca,Cr,Cwb,'bearing 1a')
bearing1b=linkages.LinearAnnularLinkage(ground,shaft1,p1b,[0,0,0],Cr,Cwb,'bearing 1b')
bearing2a=linkages.BallLinkage(ground,shaft2,p2a,[0,0,0],Ca,Cr,Cwb,'bearing 2a')
bearing2b=linkages.LinearAnnularLinkage(ground,shaft2,p2b,[0,0,0],Cr,Cwb,'bearing 2b')

dir_axis=npy.array([1,0,0])
dgs1=npy.cross(p1b-p1a,p2a-p1a)
egs1=genmechanics.geometry.Direction2Euler(dgs1,dir_axis)
gearset12=linkages.GearSetLinkage(shaft1,shaft2,pgs1,egs1,alpha_gs1,beta_gs1,Cf,Cvgs,'Gear set 1')
#gearset23=linkages.GearSetLinkage(shaft2,shaft3,[3*L/2,e1*r1+(e1+e2)*(1-r1),0],dir23)

imposed_speeds=[(bearing1a,0,w)]

load1=loads.KnownLoad(shaft1,[-L/4,0,0],[0,0,0],[0,0,0],[C,0,0],'input torque')
load2=loads.SimpleUnknownLoad(shaft2,[3*L/2,0,0],[0,0,0],[],[0],'output torque')

mech=genmechanics.Mechanism([bearing1a,bearing1b,bearing2a,bearing2b,gearset12],ground,imposed_speeds,[load1],[load2])

#r=mech.StaticAnalysis()

for l,r in mech.static_results.items():
    for d,v in r.items(): 
        print(l.name,d,v)
    
    
print('Cth: ',-r1/(1-r1)*C)


mech.GlobalSankey()


# -*- coding: utf-8 -*-
"""
Setup install script for genmechanics

"""

from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()
    
    
from os.path import dirname, isdir, join
import re
from subprocess import CalledProcessError, check_output


tag_re = re.compile(r'\btag: %s([0-9][^,]*)\b')
version_re = re.compile('^Version: (.+)$', re.M)


def get_version():
    # Return the version if it has been injected into the file by git-archive
    version = tag_re.search('$Format:%D$')
    if version:
        return version.group(1)

    d = dirname(__file__)
    
    if isdir(join(d, '.git')):
        cmd = 'git describe --tags  --dirty'
        try:
            version = check_output(cmd.split()).decode().strip()[:]
        except CalledProcessError:
            raise RuntimeError('Unable to get version number from git tags')
        if version[0]=='v':
            version = version[1:]
        # PEP 440 compatibility
        if '-' in version:
#            if version.endswith('-dirty'):
            version = '.dev'.join(version.split('-')[:2])+'-dirty'
        else:
            version = '.dev'.join(version.split('-')[:2])

    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, 'PKG-INFO')) as f:
            version = version_re.search(f.read()).group(1)
            
    # Writing to file
    with open('genmechanics/version.py', 'w') as vf:
        vf.write("# -*- coding: utf-8 -*-\nversion = '{}'".format(version))
                 
    return version



setup(name='genmechanics',
#      use_scm_version={'write_to':'genmechanics/version.py'},
#      setup_requires=['setuptools_scm'],
      version = get_version(),
      description='General mechanics solver for python',
      long_description=readme(),
      keywords='General mechanics',
      url='http://github.com/masfaraud/genmechanics',
      author='Steven Masfaraud',
      author_email='steven@masfaraud.fr',
      license='Creative Commons Attribution-Share Alike license',
      packages=['genmechanics'],
      package_dir={'genmechanics':'genmechanics'},
      install_requires=['numpy','matplotlib','networkx','scipy'],
      classifiers=['Topic :: Scientific/Engineering','Development Status :: 3 - Alpha'])
